from __future__ import annotations

import base64
import json
import logging
import os
from datetime import datetime, timedelta, timezone
from pathlib import Path

import kubernetes
import requests
import yaml
from kubernetes import config
from kubernetes.client.rest import ApiException

logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")
logger = logging.getLogger(__name__)


def label_pod_for_deletion(namespace, pod_name):
    """
    Label a Kubernetes pod for deletion by adding a 'terminate-at' annotation.

    Parameters
    ----------
    namespace : str
        The namespace of the pod.
    pod_name : str
        The name of the pod to be labeled for deletion.

    Raises
    ------
    Exception
        If there is an error labeling the pod for deletion.
    """

    # Load Kubernetes configuration
    # config.load_incluster_config()  # Use in-cluster config
    if "KUBECONFIG_LOCAL" not in os.environ:
        os.environ["KUBECONFIG_LOCAL"] = os.environ["KUBECONFIG"]
    kubeconfig = yaml.safe_load(Path(os.environ["KUBECONFIG_LOCAL"]).read_text())
    config.load_kube_config_from_dict(kubeconfig)

    # Label the pod for deletion
    body = {
        "metadata": {
            "annotations": {"terminate-at": (datetime.now(timezone.utc) + timedelta(seconds=30)).strftime("%Y-%m-%dT%H:%M:%SZ")}
        }
    }
    try:
        with kubernetes.client.ApiClient() as api_client:
            api_instance = kubernetes.client.CoreV1Api(api_client)
            api_instance.patch_namespaced_pod(name=pod_name, namespace=namespace, body=body)
            logger.info(f"Pod {pod_name} labeled for deletion")
    except Exception as e:
        logger.error(f"Error labeling pod {pod_name} for deletion: {e}")


def get_namespaces(id_token, api_urls, private_clusters=None):
    """
    Retrieves a list of unique namespaces from multiple API URLs.

    Parameters
    ----------
    id_token : str
        The ID token used for authorization when accessing public clusters.
    api_urls : list
        A list of API URLs to query for namespaces.
    private_clusters : dict, optional
        A dictionary where keys are API URLs of private clusters and values are their respective tokens. Defaults to an empty dict.

    Returns
    -------
    list
        A list of unique namespace names retrieved from the provided API URLs.
    """
    if private_clusters is None:
        private_clusters = {}
    namespace_list = []
    for api_url in api_urls:
        if api_url in private_clusters:
            token = private_clusters[api_url]
            try:
                response = requests.get(
                    api_url + "/api/v1/namespaces", headers={"Authorization": "Bearer {}".format(token)}, verify=False
                )
            except Exception:
                continue
        else:
            try:
                response = requests.get(
                    api_url + "/api/v1/namespaces", headers={"Authorization": "Bearer {}".format(id_token)}, verify=False
                )
            except Exception:
                continue
        namespaces = json.loads(response.text)
        if "items" in namespaces:
            for namespace in namespaces["items"]:
                namespace_list.append(namespace["metadata"]["name"])
    return list(set(namespace_list))


def get_cluster_status(id_token, api_urls, cluster_names, private_clusters=None):
    """
    Retrieve the status of clusters and their nodes.

    Parameters
    ----------
    id_token : str
        The ID token for authentication.
    api_urls : list
        A list of API URLs for the clusters.
    cluster_names : dict
        A dictionary mapping API URLs to cluster names.
    private_clusters : dict, optional
        A dictionary mapping private cluster API URLs to their tokens. Defaults to {}.

    Returns
    -------
    tuple
        A tuple containing:
            - node_status_dict (dict): A dictionary mapping node names to their status and schedulability.
            - cluster_dict (dict): A dictionary mapping cluster names to their node names.
    """
    if private_clusters is None:
        private_clusters = {}
    cluster_dict = {}
    node_status_dict = {}
    for api_url in api_urls:

        if api_url in private_clusters:
            token = private_clusters[api_url]
            try:
                response = requests.get(
                    api_url + "/api/v1/nodes", headers={"Authorization": "Bearer {}".format(token)}, verify=False
                )
            except Exception:

                cluster = cluster_names[api_url]
                cluster_dict[cluster] = ["Cluster API Not Reachable"]
                node_status_dict["Cluster API Not Reachable"] = ["API"]
                continue
        else:
            if api_url.endswith("None"):
                cluster = cluster_names[api_url]
                cluster_dict[cluster] = ["Cluster API Not Reachable"]
                node_status_dict["Cluster API Not Reachable"] = ["API"]
                continue
            else:
                try:
                    response = requests.get(
                        api_url + "/api/v1/nodes", headers={"Authorization": "Bearer {}".format(id_token)}, verify=False
                    )
                except Exception:
                    cluster = cluster_names[api_url]
                    cluster_dict[cluster] = ["Cluster API Not Reachable"]
                    node_status_dict["Cluster API Not Reachable"] = ["API"]
                    continue
        nodes = json.loads(response.text)

        if "items" not in nodes:
            cluster = cluster_names[api_url]
            cluster_dict[cluster] = ["Cluster API Not Reachable"]
            node_status_dict["Cluster API Not Reachable"] = ["API"]
            continue
        for node in nodes["items"]:
            node_name = node["metadata"]["name"]
            node_status_dict[node_name] = []
            cluster = cluster_names[api_url]
            if cluster in cluster_dict:
                cluster_dict[cluster].append(node_name)
            else:
                cluster_dict[cluster] = [node_name]
            for condition in node["status"]["conditions"]:
                if condition["type"] == "Ready":
                    node_status_dict[node_name].append(condition["status"])

            if "unschedulable" in node["spec"]:
                node_status_dict[node_name].append(node["spec"]["unschedulable"])
            else:
                node_status_dict[node_name].append(False)
    return node_status_dict, cluster_dict


def get_available_resources(id_token, api_urls, cluster_names, private_clusters=None):
    """
    Retrieves available GPU, CPU, and RAM resources from multiple Kubernetes clusters.

    Parameters
    ----------
    id_token : str
        The ID token for authentication.
    api_urls : list
        List of API URLs for the Kubernetes clusters.
    cluster_names : dict
        Dictionary mapping API URLs to cluster names.
    private_clusters : list, optional
        List of private clusters with their tokens. Defaults to {}.

    Returns
    -------
    tuple
        A tuple containing:
            - gpu_dict (dict): Dictionary with GPU availability information for each node.
            - cpu_dict (dict): Dictionary with CPU availability information for each node.
            - ram_dict (dict): Dictionary with RAM availability information for each node.
            - gpu_allocations (dict): Dictionary with GPU allocation details for each pod.
    """
    if private_clusters is None:
        private_clusters = {}
    gpu_dict = {}
    cpu_dict = {}
    ram_dict = {}
    gpu_allocations = {}

    for api_url in api_urls:
        cluster_name = cluster_names[api_url]
        if api_url in private_clusters:
            token = private_clusters[api_url]
            try:
                response = requests.get(
                    api_url + "/api/v1/pods", headers={"Authorization": "Bearer {}".format(token)}, verify=False
                )
                pods = json.loads(response.text)
                response = requests.get(
                    api_url + "/api/v1/nodes", headers={"Authorization": "Bearer {}".format(token)}, verify=False
                )

                nodes = json.loads(response.text)

            except Exception:
                continue

        else:
            try:
                response = requests.get(
                    api_url + "/api/v1/pods", headers={"Authorization": "Bearer {}".format(id_token)}, verify=False
                )
                pods = json.loads(response.text)
                response = requests.get(
                    api_url + "/api/v1/nodes", headers={"Authorization": "Bearer {}".format(id_token)}, verify=False
                )

                nodes = json.loads(response.text)
            except Exception:
                continue

        node_status_dict = {}

        for node in nodes["items"]:

            node_name = "{}/{}".format(cluster_name, node["metadata"]["name"])

            node_status_dict[node_name] = []

            for condition in node["status"]["conditions"]:
                if condition["type"] == "Ready":
                    node_status_dict[node_name].append(condition["status"])

            if "unschedulable" in node["spec"]:
                node_status_dict[node_name].append(node["spec"]["unschedulable"])
            else:
                node_status_dict[node_name].append(False)

        for node in nodes["items"]:

            node_name = "{}/{}".format(cluster_name, node["metadata"]["name"])

            if "nvidia.com/gpu.product" in node["metadata"]["labels"]:
                gpu_name = node["metadata"]["labels"]["nvidia.com/gpu.product"]
            else:
                gpu_name = "N/A"

            if "nvidia.com/gpu.memory" in node["metadata"]["labels"]:
                gpu_size = str(round(int(node["metadata"]["labels"]["nvidia.com/gpu.memory"]) / 1024)) + " Gi"
            else:
                gpu_size = "N/A"

            if "nvidia.com/gpu" in node["status"]["allocatable"]:
                n_gpu_allocatable = int(node["status"]["allocatable"]["nvidia.com/gpu"])
            else:
                n_gpu_allocatable = 0
            n_cpu_allocatable = int(node["status"]["allocatable"]["cpu"])
            ram_allocatable = float(int(node["status"]["allocatable"]["memory"][:-2]) / (1024.0 * 1024.0))

            n_gpu_requested = 0
            n_cpus_requested = 0
            ram_requested = 0
            for pod in pods["items"]:
                if "nodeName" not in pod["spec"]:
                    continue
                if pod["spec"]["nodeName"] != node_name.split("/")[1]:
                    continue
                containers = pod["spec"]["containers"]
                for container in containers:
                    resources = container["resources"]

                    cpu = 0
                    ram = 0
                    if "requests" in resources:
                        req = resources["requests"]
                        if "nvidia.com/gpu" in req:

                            pod_name = pod["metadata"]["name"]
                            if pod_name.startswith("jupyter"):
                                pod_name = pod_name.replace("-2d", "-").replace("-40", "@").replace("-2e", ".")[len("jupyter-") :]
                            gpu_allocations[pod_name + ", " + pod["metadata"]["namespace"]] = {
                                "node": node_name.split("/")[1],
                                "cluster": cluster_name,
                                "namespace": pod["metadata"]["namespace"],
                                "gpu": req["nvidia.com/gpu"],
                                "gpu_name": gpu_name,
                                "gpu_size": gpu_size,
                            }

                            n_gpu_requested += int(req["nvidia.com/gpu"])
                        if "cpu" in req:
                            container_cpu = req["cpu"]
                            if container_cpu[-1] == "m":
                                container_cpu = container_cpu[:-1]
                                container_cpu = int(container_cpu) / 1000.0
                            else:
                                container_cpu = int(container_cpu)
                            if container_cpu > cpu:
                                cpu = container_cpu
                        if "memory" in req:
                            if req["memory"][-2:] == "Mi":
                                container_memory = int(req["memory"][:-2]) / 1024.0
                            elif req["memory"][-2:] == "Gi":
                                container_memory = int(req["memory"][:-2])

                            elif req["memory"][-1:] == "M":
                                container_memory = int(req["memory"][:-1]) / 1024.0
                            else:

                                container_memory = int(req["memory"]) / (1024 * 1024 * 1024)
                            if container_memory > ram:
                                ram = container_memory

                    # if 'limits' in resources:
                    #    lim = resources['limits']
                    #    if 'cpu' in lim:
                    #        container_cpu = lim['cpu']
                    #        if container_cpu[-1] == 'm':
                    #            container_cpu = container_cpu[:-1]
                    #            container_cpu = int(container_cpu)/1000.
                    #        else:
                    #            container_cpu = int(container_cpu)
                    #        if container_cpu > cpu:
                    #            cpu = container_cpu
                    #    if 'memory' in lim:
                    #        if lim['memory'][-2:] == "Mi":
                    #            container_memory = int(lim['memory'][:-2]) / 1024.
                    #        elif lim['memory'][-2:] == "Gi":
                    #            container_memory = int(lim['memory'][:-2])
                    #        if container_memory > ram:
                    #            ram = container_memory
                    n_cpus_requested += cpu
                    ram_requested += ram

            gpu_dict[node_name] = []

            if node_status_dict[node_name][0] != "True":
                gpu_dict[node_name].append("NA")
                gpu_dict[node_name].append("NA")
            elif node_status_dict[node_name][1]:
                gpu_dict[node_name].append("NA")
                gpu_dict[node_name].append("NA")
            else:
                gpu_dict[node_name].append(n_gpu_allocatable - n_gpu_requested)
                gpu_dict[node_name].append(n_gpu_allocatable)
            gpu_dict[node_name].append("{}, {}".format(gpu_name, gpu_size))

            cpu_dict[node_name] = []
            cpu_dict[node_name].append(n_cpu_allocatable - n_cpus_requested)
            cpu_dict[node_name].append(n_cpu_allocatable)
            cpu_dict[node_name].append((n_cpu_allocatable - n_cpus_requested) * 100 / n_cpu_allocatable)

            ram_dict[node_name] = []
            ram_dict[node_name].append(ram_allocatable - ram_requested)
            ram_dict[node_name].append(ram_allocatable)
            ram_dict[node_name].append((ram_allocatable - ram_requested) * 100 / ram_allocatable)
    return gpu_dict, cpu_dict, ram_dict, gpu_allocations


def get_filtered_available_nodes(gpu_dict, cpu_dict, ram_dict, gpu_request, cpu_request, memory_request):
    """
    Filters and returns nodes that meet the specified GPU, CPU, and memory requirements.

    Parameters
    ----------
    gpu_dict : dict
        A dictionary where keys are node names and values are lists containing GPU information.
    cpu_dict : dict
        A dictionary where keys are node names and values are lists containing CPU information.
    ram_dict : dict
        A dictionary where keys are node names and values are lists containing RAM information.
    gpu_request : int
        The minimum number of GPUs required.
    cpu_request : float
        The minimum amount of CPU required.
    memory_request : float
        The minimum amount of memory required.

    Returns
    -------
    tuple
        Three dictionaries containing the filtered nodes and their respective GPU, CPU, and RAM information.
    """

    filtered_nodes = []
    for node in gpu_dict:
        if (
            int(gpu_dict[node][0]) >= gpu_request
            and float(cpu_dict[node][0]) >= cpu_request
            and float(ram_dict[node][0]) >= memory_request
        ):
            filtered_nodes.append(node)

    return (
        {node: gpu_dict[node] for node in filtered_nodes},
        {node: cpu_dict[node] for node in filtered_nodes},
        {node: ram_dict[node] for node in filtered_nodes},
    )


def generate_kubeconfig(id_token, user_id, namespace, cluster_id, settings):
    """
    Generates a Kubernetes configuration dictionary for a given user and cluster.

    Parameters
    ----------
    id_token : str
        The ID token for the user.
    user_id : str
        The user ID.
    namespace : str
        The Kubernetes namespace.
    cluster_id : str
        The cluster ID.
    settings : object
        An object containing various settings, including:
        - CLUSTER_NAMES (dict): A dictionary mapping cluster names to their IDs.
        - PRIVATE_CLUSTERS (dict): A dictionary of private clusters with their tokens.
        - OIDC_ISSUER_URL (str): The OIDC issuer URL.
        - OIDC_RP_CLIENT_ID (str): The OIDC client ID.
        - OIDC_RP_CLIENT_SECRET (str): The OIDC client secret.

    Returns
    -------
    dict
        A dictionary representing the Kubernetes configuration.
    """
    cluster_apis = {k: v for v, k in settings.CLUSTER_NAMES.items()}

    if cluster_apis[cluster_id] in settings.PRIVATE_CLUSTERS:
        kube_config = {
            "apiVersion": "v1",
            "kind": "Config",
            "preferences": {},
            "current-context": "MAIA/{}".format(user_id),
            "contexts": [
                {"name": "MAIA/{}".format(user_id), "context": {"user": user_id, "cluster": "MAIA", "namespace": namespace}}
            ],
            "clusters": [
                {
                    "name": "MAIA",
                    "cluster": {
                        "certificate-authority-data": "",  # settings.CLUSTER_CA,
                        "server": cluster_apis[cluster_id],
                        "insecure-skip-tls-verify": True,
                    },
                }
            ],
            "users": [{"name": user_id, "user": {"token": settings.PRIVATE_CLUSTERS[cluster_apis[cluster_id]]}}],
        }

    else:

        kube_config = {
            "apiVersion": "v1",
            "kind": "Config",
            "preferences": {},
            "current-context": "MAIA/{}".format(user_id),
            "contexts": [
                {"name": "MAIA/{}".format(user_id), "context": {"user": user_id, "cluster": "MAIA", "namespace": namespace}}
            ],
            "clusters": [
                {
                    "name": "MAIA",
                    "cluster": {
                        "certificate-authority-data": "",  # settings.CLUSTER_CA,
                        "server": cluster_apis[cluster_id],
                        "insecure-skip-tls-verify": True,
                    },
                }
            ],
            "users": [
                {
                    "name": user_id,
                    "user": {
                        "auth-provider": {
                            "config": {
                                "idp-issuer-url": settings.OIDC_ISSUER_URL,
                                "client-id": settings.OIDC_RP_CLIENT_ID,
                                "id-token": id_token,
                                "client-secret": settings.OIDC_RP_CLIENT_SECRET,
                            },
                            "name": "oidc",
                        }
                    },
                }
            ],
        }

    return kube_config


def get_namespace_details(settings, id_token, namespace, user_id, is_admin=False):
    """
    Retrieve details about the namespace including workspace applications, remote desktops, SSH ports, MONAI models,
    Orthanc instances and deployed clusters.

    Parameters
    ----------
    settings : object
        Configuration settings containing API URLs and private cluster tokens.
    id_token : str
        Identity token for authentication.
    namespace : str
        The namespace to retrieve details for.
    user_id : str
        The user ID to filter resources.
    is_admin : bool, optional
        Flag indicating if the user has admin privileges. Defaults to False.

    Returns
    -------
    tuple
        A tuple containing:
        - maia_workspace_apps (dict): Dictionary of workspace applications with their URLs.
        - remote_desktop_dict (dict): Dictionary of remote desktop URLs for users.
        - ssh_ports (dict): Dictionary of SSH ports for users.
        - monai_models (dict): Dictionary of MONAI models.
        - orthanc_list (dict): Dictionary of Orthanc instances.
        - deployed_clusters (list): List of clusters where the namespace is deployed.
    """
    maia_workspace_apps = {}
    remote_desktop_dict = {}
    orthanc_list = []
    monai_models = {}
    ssh_ports = {}
    deployed_clusters = []

    for api_url in settings.API_URL:
        if api_url in settings.PRIVATE_CLUSTERS:
            token = settings.PRIVATE_CLUSTERS[api_url]
            response = requests.get(
                api_url + "/apis/networking.k8s.io/v1/namespaces/{}/ingresses".format(namespace),
                headers={"Authorization": "Bearer {}".format(token)},
                verify=False,
            )
        else:
            response = requests.get(
                api_url + "/apis/networking.k8s.io/v1/namespaces/{}/ingresses".format(namespace),
                headers={"Authorization": "Bearer {}".format(id_token)},
                verify=False,
            )
        ingresses = json.loads(response.text)

        if api_url in settings.PRIVATE_CLUSTERS:
            token = settings.PRIVATE_CLUSTERS[api_url]
            try:
                response = requests.get(
                    api_url + "/api/v1/namespaces/{}/services".format(namespace),
                    headers={"Authorization": "Bearer {}".format(token)},
                    verify=False,
                )
            except Exception:
                continue
        else:
            try:
                response = requests.get(
                    api_url + "/api/v1/namespaces/{}/services".format(namespace),
                    headers={"Authorization": "Bearer {}".format(id_token)},
                    verify=False,
                )
            except Exception:
                continue
        services = json.loads(response.text)

        if "code" in services:
            if services["code"] == 403:
                ...

        if "items" in ingresses:
            if "items" in services:
                if len(ingresses["items"]) > 0 or len(services["items"]) > 0:
                    deployed_clusters.append(settings.CLUSTER_NAMES[api_url])

                for ingress in ingresses["items"]:
                    for rule in ingress["spec"]["rules"]:
                        if "host" not in rule:
                            rule["host"] = settings.DEFAULT_INGRESS_HOST
                        for path in rule["http"]["paths"]:
                            if path["backend"]["service"]["name"].endswith("-maia-monailabel"):
                                monai_models[path["backend"]["service"]["name"][: -len("-maia-monailabel")]] = {
                                    "monai_label": "https://" + rule["host"] + path["path"]
                                }
                            if path["backend"]["service"]["name"] == "proxy-public":
                                maia_workspace_apps["hub"] = "https://" + rule["host"] + path["path"]
                            if path["backend"]["service"]["name"] == namespace + "-orthanc-svc":

                                maia_workspace_apps["orthanc"] = "https://" + rule["host"] + path["path"]
                                maia_workspace_apps["ohif"] = "https://" + rule["host"] + path["path"] + "/ohif/"

                            if "port" in path["backend"]["service"] and "name" in path["backend"]["service"]["port"]:
                                if path["backend"]["service"]["port"]["name"] == "orthanc":
                                    orthanc_list.append(
                                        {
                                            "name": ingress["metadata"]["name"],
                                            "dicom_port": "",
                                            "url": "https://" + rule["host"] + path["path"] + "/dicom-web/",
                                        }
                                    )
                            if path["backend"]["service"]["name"] == namespace + "-mlflow-mkg" and path["path"].endswith(
                                "mlflow"
                            ):
                                maia_workspace_apps["mlflow"] = "https://" + rule["host"] + path["path"]
                            if path["backend"]["service"]["name"] == namespace + "-mlflow-mkg" and path["path"].endswith(
                                "minio-console"
                            ):
                                maia_workspace_apps["minio_console"] = "https://" + rule["host"] + path["path"]

                for service in services["items"]:
                    for port in service["spec"]["ports"]:
                        if "name" in port and port["name"] == "remote-desktop-port":
                            if "hub" in maia_workspace_apps:
                                hub_url = maia_workspace_apps["hub"]
                                user = (
                                    service["metadata"]["name"][len("jupyter-") :]
                                    .replace("-2d", "-")
                                    .replace("-40", "@")
                                    .replace("-2e", ".")
                                )
                                url = f"{hub_url}/user/{user}/proxy/80/desktop/{user}/"
                                if user_id == user or is_admin:
                                    remote_desktop_dict[user] = url

                        if "name" in port and port["name"] == "ssh":
                            # Backward compatibility
                            if service["metadata"]["name"].endswith("-ssh"):
                                user = (
                                    service["metadata"]["name"][len("jupyter-") : -len("-ssh")]
                                    .replace("-2d", "-")
                                    .replace("-40", "@")
                                    .replace("-2e", ".")
                                )
                                if user_id == user or is_admin:
                                    ssh_ports[user] = port["port"]
                            else:
                                user = (
                                    service["metadata"]["name"][len("jupyter-") :]
                                    .replace("-2d", "-")
                                    .replace("-40", "@")
                                    .replace("-2e", ".")
                                )
                                if user_id == user or is_admin:
                                    ssh_ports[user] = port["port"]

                        if "name" in port and port["name"] == "orthanc-dicom":
                            for orthanc in orthanc_list:
                                if orthanc["name"] == service["metadata"]["labels"]["app"] + "-orthanc":
                                    if service["spec"]["type"] == "NodePort":
                                        orthanc["dicom_port"] = port["nodePort"]
                                    elif service["spec"]["type"] == "LoadBalancer":
                                        orthanc["dicom_port"] = port["port"]

        if hasattr(settings, "GLOBAL_NAMESPACES") and settings.GLOBAL_NAMESPACES is not None:
            for global_namespace in settings.GLOBAL_NAMESPACES:
                if api_url in settings.PRIVATE_CLUSTERS:
                    token = settings.PRIVATE_CLUSTERS[api_url]
                    response = requests.get(
                        api_url + "/apis/networking.k8s.io/v1/namespaces/{}/ingresses".format(global_namespace),
                        headers={"Authorization": "Bearer {}".format(token)},
                        verify=False,
                    )
                else:
                    response = requests.get(
                        api_url + "/apis/networking.k8s.io/v1/namespaces/{}/ingresses".format(global_namespace),
                        headers={"Authorization": "Bearer {}".format(id_token)},
                        verify=False,
                    )
                ingresses = json.loads(response.text)

                if api_url in settings.PRIVATE_CLUSTERS:
                    token = settings.PRIVATE_CLUSTERS[api_url]
                    try:
                        response = requests.get(
                            api_url + "/api/v1/namespaces/{}/services".format(global_namespace),
                            headers={"Authorization": "Bearer {}".format(token)},
                            verify=False,
                        )
                    except Exception:
                        continue
                else:
                    try:
                        response = requests.get(
                            api_url + "/api/v1/namespaces/{}/services".format(global_namespace),
                            headers={"Authorization": "Bearer {}".format(id_token)},
                            verify=False,
                        )
                    except Exception:
                        continue
                services = json.loads(response.text)

                if "items" in ingresses:
                    if "items" in services:
                        for ingress in ingresses["items"]:
                            for rule in ingress["spec"]["rules"]:
                                if "host" not in rule:
                                    rule["host"] = settings.DEFAULT_INGRESS_HOST
                                for path in rule["http"]["paths"]:
                                    if path["backend"]["service"]["name"] == "maia-xnat":
                                        maia_workspace_apps["xnat"] = "https://" + rule["host"] + path["path"]
                                    if path["backend"]["service"]["name"] == "istio-ingressgateway":
                                        maia_workspace_apps["kubeflow"] = "https://" + rule["host"] + path["path"]

    if "hub" not in maia_workspace_apps:
        maia_workspace_apps["hub"] = "N/A"
    if "orthanc" not in maia_workspace_apps:
        maia_workspace_apps["orthanc"] = "N/A"
    if "ohif" not in maia_workspace_apps:
        maia_workspace_apps["ohif"] = "N/A"
    if "label_studio" not in maia_workspace_apps:
        maia_workspace_apps["label_studio"] = "N/A"
    if "kubeflow" not in maia_workspace_apps:
        maia_workspace_apps["kubeflow"] = "N/A"
    if "mlflow" not in maia_workspace_apps:
        maia_workspace_apps["mlflow"] = "N/A"
    if "minio_console" not in maia_workspace_apps:
        maia_workspace_apps["minio_console"] = "N/A"
    if "xnat" not in maia_workspace_apps:
        maia_workspace_apps["xnat"] = "N/A"

    return maia_workspace_apps, remote_desktop_dict, ssh_ports, monai_models, orthanc_list, deployed_clusters


def create_namespace_from_context(namespace_id):
    """
    Create a Kubernetes namespace using the provided namespace ID.

    Parameters
    ----------
    namespace_id : str
        The ID of the namespace to be created.

    Returns
    -------
    None
        This function does not return any value. It prints the API response or an exception message.

    Raises
    ------
    ApiException
        If there is an error when calling the Kubernetes CoreV1Api to create the namespace.
    """
    with kubernetes.client.ApiClient() as api_client:
        api_instance = kubernetes.client.CoreV1Api(api_client)
        body = kubernetes.client.V1Namespace(metadata=kubernetes.client.V1ObjectMeta(name=namespace_id))
        try:
            _ = api_instance.create_namespace(body)
            # print(api_response)
        except ApiException as e:
            print("Exception when calling CoreV1Api->create_namespace: \n")
            print(e)


def create_namespace(request, settings, namespace_id, cluster_id):
    """
    Creates a Kubernetes namespace using the provided request, settings, namespace ID, and cluster ID.

    Parameters
    ----------
    request : HttpRequest
        The HTTP request object containing session and user information.
    settings : Settings
        The settings object containing configuration details.
    namespace_id : str
        The ID of the namespace to be created.
    cluster_id : str
        The ID of the Kubernetes cluster where the namespace will be created.

    Returns
    -------
    None

    Raises
    ------
    ApiException
        If an error occurs while creating the namespace using the Kubernetes API.
    """
    id_token = request.session.get("oidc_id_token")
    kubeconfig_dict = generate_kubeconfig(id_token, request.user.username, "default", cluster_id, settings=settings)
    config.load_kube_config_from_dict(kubeconfig_dict)
    with open(Path("/tmp").joinpath("kubeconfig-ns"), "w") as f:
        yaml.dump(kubeconfig_dict, f)
        os.environ["KUBECONFIG"] = str(Path("/tmp").joinpath("kubeconfig-ns"))

        create_namespace_from_context(namespace_id)


def create_cifs_secret_from_context(namespace, user_id, username, password, public_key):
    """
    Create a CIFS secret in the specified Kubernetes namespace.

    Parameters
    ----------
    namespace : str
        The Kubernetes namespace where the secret will be created.
    user_id : str
        The user ID to be used in the secret name.
    username : str
        The CIFS username to be encrypted and stored in the secret.
    password : str
        The CIFS password to be encrypted and stored in the secret.
    public_key : str
        The public key used to encrypt the username and password.

    Returns
    -------
    None

    Raises
    ------
    ApiException
        If there is an error when calling the Kubernetes API to create the secret.
    """
    from MAIA.dashboard_utils import encrypt_string

    with kubernetes.client.ApiClient() as api_client:
        api_instance = kubernetes.client.CoreV1Api(api_client)
        secret = kubernetes.client.V1Secret()
        secret.metadata = kubernetes.client.V1ObjectMeta(name=f"{user_id}-cifs", namespace=namespace)

        encrypted_username = encrypt_string(public_key, username)
        encrypted_password = encrypt_string(public_key, password)

        secret.type = "fstab/cifs"
        secret.data = {
            "username": base64.b64encode(encrypted_username.encode()).decode(),
            "password": base64.b64encode(encrypted_password.encode()).decode(),
        }

        try:
            api_response = api_instance.create_namespaced_secret(namespace, secret)
            print(api_response)
        except ApiException as e:
            print("Exception when calling CoreV1Api->create_namespaced_secret: %s\n" % e)


def create_cifs_secret(request, cluster_id, settings, namespace, user_id, username, password, public_key):
    """
    Create a CIFS secret in the specified Kubernetes namespace.

    Parameters
    ----------
    request : HttpRequest
        The HTTP request object containing session and user information.
    cluster_id : str
        The ID of the Kubernetes cluster.
    settings : dict
        The settings dictionary containing configuration details.
    namespace : str
        The Kubernetes namespace where the secret will be created.
    user_id : str
        The user ID for the CIFS secret.
    username : str
        The username for the CIFS secret.
    password : str
        The password for the CIFS secret.
    public_key : str
        The public key for the CIFS secret.

    Returns
    -------
    None
    """
    id_token = request.session.get("oidc_id_token")
    kubeconfig_dict = generate_kubeconfig(id_token, request.user.username, "default", cluster_id, settings=settings)
    config.load_kube_config_from_dict(kubeconfig_dict)
    with open(Path("/tmp").joinpath("kubeconfig-ns"), "w") as f:
        yaml.dump(kubeconfig_dict, f)
        os.environ["KUBECONFIG"] = str(Path("/tmp").joinpath("kubeconfig-ns"))

        create_cifs_secret_from_context(namespace, user_id, username, password, public_key)


def create_helm_repo_secret_from_context(repo_name, helm_repo_config, argocd_namespace="argocd"):
    """
    Create a Helm repository secret in the specified Argo CD namespace using the provided Helm repository configuration.

    Parameters
    ----------
    repo_name : str
        The name of the Helm repository.
    helm_repo_config : dict
        A dictionary containing the Helm repository configuration with the following keys:
        - "username" (str): The username for the Helm repository.
        - "password" (str): The password for the Helm repository.
        - "project" (str): The project associated with the Helm repository.
        - "url" (str): The URL of the Helm repository.
        - "type" (str): The type of the Helm repository.
        - "name" (str): The name of the Helm repository.
        - "enableOCI" (str): A flag indicating whether OCI is enabled for the Helm repository.
    argocd_namespace : str, optional
        The namespace in which to create the secret (default is "argocd").

    Returns
    -------
    None

    Raises
    ------
    ApiException
        If there is an error when calling the Kubernetes API to create the secret.
    """

    username = helm_repo_config["username"]
    password = helm_repo_config["password"]
    if "project" not in helm_repo_config:
        project = None
    else:
        project = helm_repo_config["project"]
    url = helm_repo_config["url"]
    type = helm_repo_config["type"]
    name = helm_repo_config["name"]
    enable_oci = helm_repo_config["enableOCI"]

    with kubernetes.client.ApiClient() as api_client:
        api_instance = kubernetes.client.CoreV1Api(api_client)
        secret = kubernetes.client.V1Secret()
        secret.metadata = kubernetes.client.V1ObjectMeta(
            name=f"repo-{repo_name}", labels={"argocd.argoproj.io/secret-type": "repository"}, namespace=argocd_namespace
        )
        if project is not None:
            secret.data = {
                "username": base64.b64encode(username.encode()).decode(),
                "password": base64.b64encode(password.encode()).decode(),
                "project": base64.b64encode(project.encode()).decode(),
                "url": base64.b64encode(url.encode()).decode(),
                "type": base64.b64encode(type.encode()).decode(),
                "name": base64.b64encode(name.encode()).decode(),
                "enableOCI": base64.b64encode(enable_oci.encode()).decode(),
            }
        else:
            secret.data = {
                "username": base64.b64encode(username.encode()).decode(),
                "password": base64.b64encode(password.encode()).decode(),
                "url": base64.b64encode(url.encode()).decode(),
                "type": base64.b64encode(type.encode()).decode(),
                "name": base64.b64encode(name.encode()).decode(),
                "enableOCI": base64.b64encode(enable_oci.encode()).decode(),
            }

        try:
            api_response = api_instance.create_namespaced_secret(argocd_namespace, secret)
            print(api_response)
        except ApiException as e:
            print("Exception when calling CoreV1Api->create_namespaced_secret: %s\n" % e)


def create_docker_registry_secret_from_context(docker_credentials, namespace, secret_name):
    """
    Creates a Kubernetes secret of type `kubernetes.io/dockerconfigjson` in the specified namespace
    using the provided Docker registry credentials.

    Parameters
    ----------
    docker_credentials : dict
        A dictionary containing Docker registry credentials with the following keys:
        - "registry" : str
            The Docker registry URL (e.g., "https://index.docker.io/v1/").
        - "username" : str
            The username for the Docker registry.
        - "password" : str
            The password for the Docker registry.
    namespace : str
        The Kubernetes namespace where the secret will be created.
    secret_name : str
        The name of the Kubernetes secret to be created.

    Raises
    ------
    ApiException
        If there is an error while creating the Kubernetes secret, an exception is raised
        with details about the failure.

    Notes
    -----
    This function uses the Kubernetes Python client to create the secret. Ensure that the
    Kubernetes client is properly configured to interact with the desired cluster.

    """

    docker_credentials_dict = {
        "auths": {
            docker_credentials["registry"]: {
                "username": docker_credentials["username"],
                "password": docker_credentials["password"],
                "auth": base64.b64encode(f"{docker_credentials['username']}:{docker_credentials['password']}".encode()).decode(),
            }
        }
    }
    with kubernetes.client.ApiClient() as api_client:
        api_instance = kubernetes.client.CoreV1Api(api_client)
        secret = kubernetes.client.V1Secret()
        secret.metadata = kubernetes.client.V1ObjectMeta(name=secret_name, namespace=namespace)
        secret.data = {
            ".dockerconfigjson": base64.b64encode(json.dumps(docker_credentials_dict).encode()).decode(),
            "config.json": base64.b64encode(
                json.dumps(docker_credentials_dict).encode()
            ).decode(),  # Needed for KubeFlow to mount the secret with Kaniko
        }
        secret.type = "kubernetes.io/dockerconfigjson"

        try:
            api_response = api_instance.create_namespaced_secret(namespace, secret)
            print(api_response)
        except ApiException as e:
            print("Exception when calling CoreV1Api->create_namespaced_secret: %s\n" % e)


def retrieve_json_key_for_maia_registry_authentication_from_context(namespace, secret_name, registry_url):
    """
    Retrieve the JSON key for MAIA registry authentication from a Kubernetes secret.
    This function reads a Kubernetes secret in the specified namespace, decodes the `.dockerconfigjson`
    field, and extracts the password for the given registry URL.

    Parameters
    ----------
    namespace : str
        The namespace in which the Kubernetes secret is located.
    secret_name : str
        The name of the Kubernetes secret containing the `.dockerconfigjson`.
    registry_url : str
        The URL of the container registry for which the authentication key is required.

    Returns
    -------
    str
        The password associated with the specified registry URL in the `.dockerconfigjson`.
        Returns an empty dictionary if an exception occurs.

    Raises
    ------
    kubernetes.client.exceptions.ApiException
        If there is an error while reading the Kubernetes secret.
    """
    with kubernetes.client.ApiClient() as api_client:
        api_instance = kubernetes.client.CoreV1Api(api_client)

        try:
            secret = api_instance.read_namespaced_secret(name=secret_name, namespace=namespace)

            # Decode the .dockerconfigjson
            dockerconfig = base64.b64decode(secret.data[".dockerconfigjson"]).decode()
            dockerconfig_json = json.loads(dockerconfig)
            return dockerconfig_json["auths"][registry_url]["password"]
        except ApiException as e:
            print("Exception when calling CoreV1Api->read_namespaced_secret: %s\n" % e)
            return {}


def retrieve_json_key_for_maia_registry_authentication(request, cluster_id, settings, namespace, secret_name, registry_url):
    """
    Retrieves the JSON key for MAIA registry authentication.

    This function generates a kubeconfig dictionary using the provided
    OpenID Connect (OIDC) ID token and user information, writes it to a
    temporary kubeconfig file, and sets the `KUBECONFIG` environment
    variable. It then delegates the retrieval of the JSON key to another
    function.

    Parameters
    ----------
    request : HttpRequest
        The HTTP request object containing the session and user information.
    cluster_id : str
        The ID of the Kubernetes cluster.
    settings : dict
        The settings dictionary containing configuration details.
    namespace : str
        The Kubernetes namespace where the secret is located.
    secret_name : str
        The name of the Kubernetes secret containing the registry credentials.
    registry_url : str
        The URL of the container registry.

    Returns
    -------
    dict
        A dictionary containing the JSON key for MAIA registry authentication.

    Raises
    ------
    KeyError
        If the `oidc_id_token` is not found in the session.
    FileNotFoundError
        If there is an issue writing the kubeconfig file to the temporary directory.
    """
    id_token = request.session.get("oidc_id_token")
    kubeconfig_dict = generate_kubeconfig(id_token, request.user.username, "default", cluster_id, settings=settings)
    config.load_kube_config_from_dict(kubeconfig_dict)
    with open(Path("/tmp").joinpath("kubeconfig-ns"), "w") as f:
        yaml.dump(kubeconfig_dict, f)
        os.environ["KUBECONFIG"] = str(Path("/tmp").joinpath("kubeconfig-ns"))

        return retrieve_json_key_for_maia_registry_authentication_from_context(namespace, secret_name, registry_url)



from __future__ import annotations

import os
from pathlib import Path

import yaml


def deploy_maia_kaniko(
    namespace,
    config_folder,
    cluster_config_dict,
    release_name,
    project_id,
    registry_url,
    registry_secret_name,
    image_name,
    image_tag,
    subpath,
    build_args=None,
    registry_credentials=None,
):
    """
    Deploys a Kaniko job for building and pushing Docker images to a specified registry.

    Parameters
    ----------
    namespace : str
        The Kubernetes namespace where the Kaniko job will be deployed.
    config_folder : str
        The folder path where the configuration files will be stored.
    cluster_config_dict : dict
        Dictionary containing cluster configuration details, including storage class.
    release_name : str
        The release name for the Kaniko job.
    project_id : str
        The project identifier.
    registry_url : str
        The URL of the Docker registry where the image will be pushed.
    registry_secret_name : str
        The name of the Kubernetes secret for accessing the Docker registry.
    image_name : str
        The name of the Docker image to be built.
    image_tag : str
        The tag of the Docker image to be built.
    subpath : str
        The subpath of the repository where the Dockerfile is located.
    build_args : list, optional
        A list of build arguments to be passed to the Kaniko job.
    registry_credentials : dict, optional
        A dictionary containing registry credentials with keys 'username', 'password', 'server', and 'email'.

    Returns
    -------
    dict
        A dictionary containing deployment details including namespace, release name,
        chart name, repo URL, chart version, and values file path.
    """

    if "MAIA_HELM_REPO_URL" not in os.environ:
        raise ValueError(
            "MAIA_HELM_REPO_URL environment variable not set. Please set this variable to the URL of the MAIA Helm repository. Example: https://kthcloud.github.io/MAIA/"  # noqa: B950
        )

    kaniko_values = {
        "chart_name": "mkg-kaniko",
        "repo_url": os.environ["MAIA_HELM_REPO_URL"],
        "chart_version": "1.0.3",
        "namespace": "mkg-kaniko",
    }

    if "MAIA_GIT_REPO_URL" not in os.environ:
        raise ValueError(
            "MAIA_GIT_REPO_URL environment variable not set. Please set this variable to the URL of the MAIA Git repository with the Docker Image to build. Example: git://github.com/kthcloud/MAIA.git"  # noqa: B950
        )

    if "GIT_USERNAME" not in os.environ:
        raise ValueError(
            "GIT_USERNAME environment variable not set. Please set this variable to the username for accessing the Git repository with the Docker Image to build."  # noqa: B950
        )

    if "GIT_TOKEN" not in os.environ:
        raise ValueError(
            "GIT_TOKEN environment variable not set. Please set this variable to the personal access token for accessing the Git repository with the Docker Image to build."  # noqa: B950
        )

    registry_server = registry_credentials.get("server") if registry_credentials else None
    registry_username = registry_credentials.get("username") if registry_credentials else None
    registry_password = registry_credentials.get("password") if registry_credentials else None
    registry_email = registry_credentials.get("email") if registry_credentials else None

    kaniko_values.update(
        {
            "docker_registry_secret": registry_secret_name,
            "namespace": "mkg-kaniko",
            "dockerRegistryServer": (
                "https://" + registry_server if "registry_server" not in cluster_config_dict else registry_server
            ),
            "dockerRegistryUsername": registry_username,
            "dockerRegistryPassword": registry_password,
            "dockerRegistryEmail": registry_email,
            "dockerRegistrySecretName": registry_secret_name,
            "pvc": {"pvc_type": cluster_config_dict["storage_class"], "size": "10Gi"},
            "git_username": os.environ.get("GIT_USERNAME"),
            "git_token": os.environ.get("GIT_TOKEN"),
            "args": [
                "--dockerfile=Dockerfile",
                f"--context={os.environ['MAIA_GIT_REPO_URL']}",  # refs/heads/mybranch
                "--context-sub-path=" + subpath,
                "--destination={}/{}:{}".format(registry_url, image_name, image_tag),
                "--cache=true",
                "--cache-dir=/cache",
            ],
        }
    )

    if build_args:
        for build_arg in build_args:
            kaniko_values["args"].append(f"--build-arg={build_arg}")

    release_name_values = release_name.replace("-", "_")
    Path(config_folder).joinpath(project_id, f"{release_name_values}_values").mkdir(parents=True, exist_ok=True)
    with open(
        Path(config_folder).joinpath(project_id, f"{release_name_values}_values", f"{release_name_values}_values.yaml"), "w"
    ) as f:
        yaml.dump(kaniko_values, f)

    return {
        "namespace": namespace,
        "release": f"{project_id}-{release_name}",
        "chart": kaniko_values["chart_name"],
        "repo": kaniko_values["repo_url"],
        "version": kaniko_values["chart_version"],
        "values": str(
            Path(config_folder).joinpath(project_id, f"{release_name_values}_values", f"{release_name_values}_values.yaml")
        ),
    }



#!/usr/bin/env python
from __future__ import annotations

import asyncio
import datetime
import os
from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from textwrap import dedent

import click
import yaml
from hydra import compose as hydra_compose
from hydra import initialize_config_dir
from omegaconf import OmegaConf
from pyhelm3 import Client

import MAIA
from MAIA.maia_admin import (
    create_harbor_values,
    create_keycloak_values,
    create_maia_admin_toolkit_values,
    create_maia_dashboard_values,
    install_maia_project,
)
from MAIA.maia_core import create_rancher_values

version = MAIA.__version__


TIMESTAMP = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.datetime.now())

DESC = dedent(
    """
    Script to Install MAIA Admin Toolkit to a Kubernetes cluster from ArgoCD. The specific MAIA configuration
    is specified by setting the corresponding ``--maia-config-file``, and the cluster configuration is
    specified by setting the corresponding ``--cluster-config``.
    """  # noqa: E501
)
EPILOG = dedent(
    """
    Example call:
    ::
        {filename} --maia-config-file /PATH/TO/maia_config.yaml --cluster-config /PATH/TO/clusyer_config.yaml
        --config-folder /PATH/TO/config_folder
    """.format(  # noqa: E501
        filename=Path(__file__).stem
    )
)


def get_arg_parser():
    pars = ArgumentParser(description=DESC, epilog=EPILOG, formatter_class=RawTextHelpFormatter)

    pars.add_argument(
        "--maia-config-file", type=str, required=True, help="YAML configuration file used to extract MAIA configuration."
    )

    pars.add_argument(
        "--cluster-config", type=str, required=True, help="YAML configuration file used to extract the cluster configuration."
    )

    pars.add_argument(
        "--config-folder",
        type=str,
        required=True,
        help="Configuration Folder where to locate (and temporarily store) the MAIA configuration files.",
    )

    pars.add_argument("-v", "--version", action="version", version="%(prog)s " + version)

    return pars


async def verify_installed_maia_admin_toolkit(project_id, namespace):

    print(os.environ["KUBECONFIG"])
    client = Client(kubeconfig=os.environ["KUBECONFIG"])

    try:
        revision = await client.get_current_revision(project_id, namespace=namespace)
    except Exception as e:
        print(f"An error occurred: {e}")
        return -1
    chart_metadata = await revision.chart_metadata()
    print(
        revision.release.name,
        revision.release.namespace,
        revision.revision,
        str(revision.status),
        chart_metadata.name,
        chart_metadata.version,
    )
    return revision.revision


@click.command()
@click.option("--maia-config-file", type=str)
@click.option("--cluster-config", type=str)
@click.option("--config-folder", type=str)
def main(maia_config_file, cluster_config, config_folder):
    install_maia_admin_toolkit(maia_config_file, cluster_config, config_folder)


def install_maia_admin_toolkit(maia_config_file, cluster_config, config_folder):

    maia_config_dict = yaml.safe_load(Path(maia_config_file).read_text())

    cluster_config_dict = yaml.safe_load(Path(cluster_config).read_text())

    admin_group_id = maia_config_dict["admin_group_ID"]
    project_id = "maia-admin"

    cluster_address = "https://kubernetes.default.svc"  # TODO: Change this to make it configurable

    helm_commands = []

    helm_commands.append(create_harbor_values(config_folder, project_id, cluster_config_dict))
    helm_commands.append(create_keycloak_values(config_folder, project_id, cluster_config_dict))
    helm_commands.append(create_rancher_values(config_folder, project_id, cluster_config_dict))
    helm_commands.append(
        create_maia_admin_toolkit_values(config_folder, project_id, cluster_config_dict, maia_config_dict=maia_config_dict)
    )
    helm_commands.append(
        create_maia_dashboard_values(config_folder, project_id, cluster_config_dict, maia_config_dict=maia_config_dict)
    )

    for helm_command in helm_commands:
        cmd = [
            "helm",
            "upgrade",
            "--install",
            "--wait",
            "-n",
            helm_command["namespace"],
            helm_command["release"],
            helm_command["chart"],
            "--repo",
            helm_command["repo"],
            "--version",
            helm_command["version"],
            "--values",
            helm_command["values"],
        ]
        print(" ".join(cmd))

    values = {
        "defaults": [
            "_self_",
            {"harbor_values": "harbor_values"},
            {"keycloak_values": "keycloak_values"},
            {"maia_admin_toolkit_values": "maia_admin_toolkit_values"},
            {"maia_dashboard_values": "maia_dashboard_values"},
            {"rancher_values": "rancher_values"},
        ],
        "argo_namespace": maia_config_dict["argocd_namespace"],
        "admin_group_ID": admin_group_id,
        "destination_server": f"{cluster_address}",
        "sourceRepos": [
            "https://kthcloud.github.io/MAIA/",
            "https://helm.goharbor.io",
            "https://charts.bitnami.com/bitnami",
            "https://releases.rancher.com/server-charts/latest",
        ],
    }
    Path(config_folder).joinpath(project_id).mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id, "values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(values))

    initialize_config_dir(config_dir=str(Path(config_folder).joinpath(project_id)), job_name=project_id)
    cfg = hydra_compose("values.yaml")
    OmegaConf.save(cfg, str(Path(config_folder).joinpath(project_id, f"{project_id}_values.yaml")), resolve=True)

    revision = asyncio.run(verify_installed_maia_admin_toolkit(project_id, maia_config_dict["argocd_namespace"]))

    if revision == -1:
        print("Installing MAIA Admin Toolkit")

        project_chart = maia_config_dict["admin_project_chart"]
        project_repo = maia_config_dict["admin_project_repo"]
        project_version = maia_config_dict["admin_project_version"]
        asyncio.run(
            install_maia_project(
                project_id,
                Path(config_folder).joinpath(project_id, f"{project_id}_values.yaml"),
                maia_config_dict["argocd_namespace"],
                project_chart,
                project_repo=project_repo,
                project_version=project_version,
            )
        )
    else:
        print("Upgrading MAIA Admin Toolkit")

        project_chart = maia_config_dict["admin_project_chart"]
        project_repo = maia_config_dict["admin_project_repo"]
        project_version = maia_config_dict["admin_project_version"]
        asyncio.run(
            install_maia_project(
                project_id,
                Path(config_folder).joinpath(project_id, f"{project_id}_values.yaml"),
                maia_config_dict["argocd_namespace"],
                project_chart,
                project_repo=project_repo,
                project_version=project_version,
            )
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import annotations

import argparse
import datetime
import json
import os
import subprocess
from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from textwrap import dedent

import yaml

import MAIA
from MAIA.helm_values import read_config_dict_and_generate_helm_values_dict

version = MAIA.__version__

TIMESTAMP = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.datetime.now())

DESC = dedent(
    """
    Script to deploy the MAIAKubeGate helm chart to a Kubernetes cluster. The target cluster is specified by setting ``KUBECONFIG``
    as an environment variable, while the configuration file for the chart is specified with ``config_file``.
    """  # noqa: E501
)
EPILOG = dedent(
    """
    Example call:
    ::
        {filename}  --config-file /PATH/TO/config_file.json
    """.format(  # noqa: E501
        filename=Path(__file__).stem
    )
)


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def get_arg_parser():
    pars = ArgumentParser(description=DESC, epilog=EPILOG, formatter_class=RawTextHelpFormatter)

    pars.add_argument(
        "--config-file",
        type=str,
        required=True,
        help="JSON configuration file used to generate the custom values used in the Helm chart.",
    )

    pars.add_argument(
        "--print-helm-values-only",
        type=str,
        required=False,
        help="Flag to save the generated Helm values on the specified file and exit.",
    )

    pars.add_argument("-v", "--version", action="version", version="%(prog)s " + version)

    return pars


def main():
    parser = get_arg_parser()

    arguments = vars(parser.parse_args())

    config_file = arguments["config_file"]

    with open(config_file) as json_file:
        config_dict = json.load(json_file)

    kubeconfig = yaml.safe_load(Path(os.environ["KUBECONFIG"]).read_text())

    ssh_process = subprocess.Popen(["sh"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True, bufsize=0)

    helm_dict = read_config_dict_and_generate_helm_values_dict(config_dict, kubeconfig)

    if arguments["print_helm_values_only"]:
        print(helm_dict)
        with open(arguments["print_helm_values_only"], "w") as f:
            yaml.dump(helm_dict, f)
        return

    chart_name = config_dict["chart_name"]
    with open(f"./{chart_name}_values.yaml", "w") as f:  # TODO: remove this and load values from memory
        yaml.dump(helm_dict, f)

    ssh_process.stdin.write(
        "helm upgrade --install {} --namespace={} maia/mkg --values ./{}_values.yaml\n".format(
            config_dict["chart_name"], config_dict["namespace"], config_dict["chart_name"]
        )
    )

    ssh_process.stdin.close()
    for line in ssh_process.stdout:
        if line == "END\n":
            break
        print(line, end="")


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
from pathlib import Path

from kubernetes import client, config
from omegaconf import OmegaConf


def create_prometheus_values(config_folder, project_id, cluster_config_dict, maia_config_dict):
    """
    Generates Prometheus values configuration for a Kubernetes cluster and writes it to a YAML file.

    Parameters
    ----------
    config_folder : str
        The folder where the configuration files will be stored.
    project_id : str
        The project identifier.
    cluster_config_dict : dict
        Dictionary containing cluster configuration details.
    maia_config_dict : dict
        Dictionary containing MAIA configuration details.

    Returns
    -------
    dict
        A dictionary containing the namespace, repository URL, chart version, path to the values file, release name, and chart name.
    """

    config.load_kube_config(config_file=os.environ.get("DEPLOY_KUBECONFIG", None))

    # Create a CoreV1Api instance
    v1 = client.CoreV1Api()

    # Get the list of nodes
    nodes = v1.list_node()

    # Extract InternalIP for each node
    internal_ips = []
    for node in nodes.items:
        for address in node.status.addresses:
            if address.type == "InternalIP":
                internal_ips.append(address.address)

    prometheus_values = {
        "namespace": "observability",
        "chart_version": "45.5.0",
        "repo_url": "https://prometheus-community.github.io/helm-charts",
        "chart_name": "kube-prometheus-stack",
    }  # TODO: Change this to updated values

    admin_group_id = maia_config_dict["admin_group_ID"]
    domain = cluster_config_dict["domain"]
    prometheus_values.update(
        {
            "kubeControllerManager": {"endpoints": internal_ips},
            "kubeScheduler": {"endpoints": internal_ips},
            "grafana": {
                "grafana.ini": {
                    "server": {"root_url": "https://grafana." + cluster_config_dict["domain"]},
                    "auth.generic_oauth": {
                        "api_url": f"https://iam.{domain}/realms/maia/protocol/openid-connect/userinfo",
                        "auth_url": f"https://iam.{domain}/realms/maia/protocol/openid-connect/auth",
                        "client_id": "maia",
                        "client_secret": cluster_config_dict["keycloak"]["client_secret"],
                        "enabled": True,
                        "name": "OAuth",
                        "empty_scopes": False,
                        "role_attribute_path": f"contains(groups[*], '{admin_group_id}') && 'Admin' || 'Viewer'",
                        "scopes": "openid profile email",
                        "team_ids": admin_group_id,
                        "team_ids_attribute_path": "groups[*]",
                        "teams_url": f"https://iam.{domain}/realms/maia/protocol/openid-connect/userinfo",
                        "token_url": f"https://iam.{domain}/realms/maia/protocol/openid-connect/token",
                    },
                },
                "assertNoLeakedSecrets": False,
                "defaultDashboardsEnabled": True,
                "persistence": {"enabled": True},
                "ingress": {
                    "enabled": True,
                    "annotations": {},
                    "hosts": ["grafana." + cluster_config_dict["domain"]],
                    "tls": [{"hosts": ["grafana." + cluster_config_dict["domain"]]}],
                },
                "additionalDataSources": [
                    {"name": "loki", "type": "loki", "url": "http://maia-core-loki.observability.svc.cluster.local:3100"},
                    {"name": "tempo", "type": "tempo", "url": "http://maia-core-tempo.observability.svc.cluster.local:3100"},
                ],
            },
        }
    )

    gpu_enabled = True

    if gpu_enabled:
        prometheus_values["prometheus"] = {
            "prometheusSpec": {
                "additionalScrapeConfigs": [
                    {
                        "job_name": "gpu-metrics",
                        "kubernetes_sd_configs": [{"namespaces": {"names": ["gpu-operator"]}, "role": "endpoints"}],
                        "metrics_path": "/metrics",
                        "relabel_configs": [
                            {
                                "action": "replace",
                                "source_labels": ["__meta_kubernetes_pod_node_name"],
                                "target_label": "kubernetes_node",
                            }
                        ],
                        "scheme": "http",
                        "scrape_interval": "1s",
                    }
                ],
                "serviceMonitorSelectorNilUsesHelmValues": False,
            }
        }

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        prometheus_values["grafana"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        prometheus_values["grafana"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = "true"
        prometheus_values["grafana"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = (
            cluster_config_dict["traefik_resolver"]
        )
    elif cluster_config_dict["ingress_class"] == "nginx":
        prometheus_values["grafana"]["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"
        prometheus_values["grafana"]["ingress"]["tls"][0]["secretName"] = "grafana." + cluster_config_dict["domain"]

    Path(config_folder).joinpath(project_id, "prometheus_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "prometheus_values", "prometheus_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(prometheus_values))

    return {
        "namespace": prometheus_values["namespace"],
        "repo": prometheus_values["repo_url"],
        "version": prometheus_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "prometheus_values", "prometheus_values.yaml")),
        "release": f"{project_id}-prometheus",
        "chart": prometheus_values["chart_name"],
    }


def create_loki_values(config_folder, project_id):
    """
    Creates and writes Loki values configuration to a YAML file and returns deployment details.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The project identifier.

    Returns
    -------
    dict
        A dictionary containing deployment details including namespace, repo URL, chart version,
        values file path, release name, and chart name.
    """

    loki_values = {
        "namespace": "observability",
        "chart_version": "2.9.9",
        "repo_url": "https://grafana.github.io/helm-charts",
        "chart_name": "loki-stack",
    }  # TODO: Change this to updated values

    loki_values.update({"grafana": {"sidecar": {"datasources": {"enabled": False}}}})

    Path(config_folder).joinpath(project_id, "loki_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "loki_values", "loki_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(loki_values))

    return {
        "namespace": loki_values["namespace"],
        "repo": loki_values["repo_url"],
        "version": loki_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "loki_values", "loki_values.yaml")),
        "release": f"{project_id}-loki",
        "chart": loki_values["chart_name"],
    }


def create_tempo_values(config_folder, project_id):
    """
    Creates a set of tempo values and writes them to a YAML file in the specified configuration folder.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder where the tempo values will be stored.
    project_id : str
        The project identifier used to create a subdirectory and name the YAML file.

    Returns
    -------
    dict
        A dictionary containing the following keys:
        - namespace (str): The namespace for the tempo values.
        - repo (str): The repository URL for the Helm chart.
        - version (str): The version of the Helm chart.
        - values (str): The path to the generated tempo values YAML file.
        - release (str): The release name for the Helm chart.
        - chart (str): The name of the Helm chart.
    """

    tempo_values = {
        "namespace": "observability",
        "chart_version": "1.0.0",
        "repo_url": "https://grafana.github.io/helm-charts",
        "chart_name": "tempo",
    }  # TODO: Change this to updated values

    Path(config_folder).joinpath(project_id, "tempo_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "tempo_values", "tempo_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(tempo_values))

    return {
        "namespace": tempo_values["namespace"],
        "repo": tempo_values["repo_url"],
        "version": tempo_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "tempo_values", "tempo_values.yaml")),
        "release": f"{project_id}-tempo",
        "chart": tempo_values["chart_name"],
    }


def create_core_toolkit_values(config_folder, project_id, cluster_config_dict):
    """
    Creates and saves the core toolkit values for a Kubernetes cluster.
    This function generates a dictionary of core toolkit values based on the provided
    configuration folder, project ID, and cluster configuration dictionary. It retrieves
    the internal IP addresses of the nodes in the Kubernetes cluster and uses them to
    configure the MetalLB addresses. The generated values are saved to a YAML file.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The project identifier.
    cluster_config_dict : dict
        A dictionary containing cluster configuration details, including 'ingress_class' and 'ingress_resolver_email'.

    Returns
    -------
    dict
        A dictionary containing the namespace, repository URL, chart version, path to the
        values YAML file, release name, and chart name.
    """

    config.load_kube_config()

    # Create a CoreV1Api instance
    v1 = client.CoreV1Api()

    # Get the list of nodes
    nodes = v1.list_node()

    # Extract InternalIP for each node
    internal_ips = []
    for node in nodes.items:
        for address in node.status.addresses:
            if address.type == "InternalIP":
                internal_ips.append(address.address)

    if len(internal_ips) == 1:
        internal_ips.append(internal_ips[0])

    core_toolkit_values = {
        "namespace": "maia-core-toolkit",
        "chart_version": "0.1.7",
        "repo_url": os.environ.get("MAIA_PRIVATE_REGISTRY", None),
        "chart_name": "maia-core-toolkit",
    }  # TODO: Change this to updated values

    core_toolkit_values.update(
        {
            "metric_server": {"enabled": True},  # Replace with https://artifacthub.io/packages/helm/metrics-server/metrics-server
            "dashboard": {"enabled": True},
            "default_ingress_class": cluster_config_dict["ingress_class"],
            "cert_manager": {"enabled": True, "email": cluster_config_dict["ingress_resolver_email"]},
            "metallb": {"enabled": True, "addresses": "-".join(internal_ips)},
        }
    )

    Path(config_folder).joinpath(project_id, "core_toolkit_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "core_toolkit_values", "core_toolkit_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(core_toolkit_values))

    return {
        "namespace": core_toolkit_values["namespace"],
        "repo": core_toolkit_values["repo_url"],
        "version": core_toolkit_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "core_toolkit_values", "core_toolkit_values.yaml")),
        "release": f"{project_id}-toolkit",
        "chart": core_toolkit_values["chart_name"],
    }


def create_traefik_values(config_folder, project_id, cluster_config_dict):
    """
    Creates the Traefik values configuration file for a given project and cluster configuration.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The unique identifier for the project.
    cluster_config_dict : dict
        A dictionary containing the cluster configuration.

    Returns
    -------
    dict
        A dictionary containing the namespace, repository URL, chart version, values file path, release name,
        and chart name for the Traefik deployment.

    Raises
    ------
    OSError
        If there is an error creating the directory or writing the file.

    Example
    -------
    config_folder = "/path/to/config"
    project_id = "my_project"
    cluster_config_dict = {
        "traefik_resolver": "myresolver",
        "ingress_resolver_email": "email@example.com",
        "domain": "example.com",
        "traefik_dashboard_password": "password"
    create_traefik_values(config_folder, project_id, cluster_config_dict)
    """
    Path(config_folder).joinpath(project_id, "traefik_values").mkdir(parents=True, exist_ok=True)

    self_signed_tls = False
    traefik_values = {
        "namespace": "traefik",
        "repo_url": "https://traefik.github.io/charts",
        "chart_name": "traefik",
        "chart_version": "33.2.1",
    }  # TODO: Change this to updated values

    traefik_values.update(
        {
            "certificatesResolvers": {
                "maiaresolver": {
                    "acme": {
                        "email": cluster_config_dict["ingress_resolver_email"],
                        # "httpchallenge": "true",
                        "httpchallenge": {"entryPoint": "web"},
                        "caserver": "https://acme-v02.api.letsencrypt.org/directory",
                        # "caServer": "https://acme-staging-v02.api.letsencrypt.org/directory",
                        "storage": "/data/acme.json",
                    }
                }
            },
            "ingressRoute": {
                "dashboard": {
                    "enabled": True,
                    "matchRule": "Host(`{}`)".format("traefik." + cluster_config_dict["domain"]),  # && PathPrefix(`/dashboard`)
                    "entryPoints": ["websecure"],
                    "tls": {"certResolver": cluster_config_dict["traefik_resolver"]},
                    "middlewares": [{"name": "traefik-dashboard-auth"}, {"name": "traefik-dashboard-replace-path"}],
                }
            },
            "extraObjects": [
                {
                    "apiVersion": "v1",
                    "kind": "Secret",
                    "metadata": {"name": "traefik-dashboard-auth-secret"},
                    "type": "kubernetes.io/basic-auth",
                    "stringData": {"username": "admin", "password": cluster_config_dict["traefik_dashboard_password"]},
                },
                {
                    "apiVersion": "traefik.io/v1alpha1",
                    "kind": "Middleware",
                    "metadata": {"name": "traefik-dashboard-replace-path"},
                    "spec": {"replacePathRegex": {"regex": "^/dashboard/(.*)", "replacement": "/dashboard/$1"}},
                },
                {
                    "apiVersion": "traefik.io/v1alpha1",
                    "kind": "Middleware",
                    "metadata": {"name": "traefik-dashboard-auth"},
                    "spec": {"basicAuth": {"secret": "traefik-dashboard-auth-secret"}},
                },
            ],
            "globalArguments": [
                "--entrypoints.web.http.redirections.entryPoint.to=:443",
                # "--entrypoints.web.http.redirections.entrypoint.to=websecure",
                "--entrypoints.web.http.redirections.entrypoint.scheme=https",
                "--global.checknewversion",
                "--global.sendanonymoususage",
                # "--entrypoints.metrics.address=:9100/tcp",
                # "--entrypoints.traefik.address=:9000/tcp",
                "--api.dashboard=true",
                "--ping=true",
                "--metrics.prometheus=true",
                "--metrics.prometheus.entrypoint=metrics",
                "--providers.kubernetescrd",
                "--providers.kubernetesingress",
                # "--entrypoints.websecure.http.tls=true",
                "--api.insecure",
                "--accesslog",
                # "--entrypoints.web.Address=:80",
                "--entryPoints.websecure.forwardedHeaders.insecure",
                "--entryPoints.web.forwardedHeaders.insecure",
                # "--entrypoints.websecure.Address=:443",
                # f"--certificatesresolvers.{resolver}.acme.httpchallenge=true",
                # f"--certificatesresolvers.{resolver}).acme.httpchallenge.entrypoint=web",
                # f"--certificatesresolvers.{resolver}.acme.email={acme_email}",
                # f"--certificatesresolvers.{resolver}.acme.storage=/data/acme.json",
                # f"--certificatesresolvers.{resolver}.acme.caserver=https://acme-v02.api.letsencrypt.org/directory"
            ],
        }
    )

    if self_signed_tls:
        traefik_values.update({"tlsStore": {"default": {"defaultCertificate": {"secretName": "wildcard-domain-tls"}}}})

    with open(Path(config_folder).joinpath(project_id, "traefik_values", "traefik_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(traefik_values))

    return {
        "namespace": traefik_values["namespace"],
        "repo": traefik_values["repo_url"],
        "version": traefik_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "traefik_values", "traefik_values.yaml")),
        "release": f"{project_id}-traefik",
        "chart": traefik_values["chart_name"],
    }


def create_metallb_values(config_folder, project_id):
    """
    Creates and writes MetalLB Helm chart values to a YAML file and returns a dictionary with deployment details.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder where the YAML file will be created.
    project_id : str
        The project identifier used to create a unique directory and release name.

    Returns
    -------
    dict
        A dictionary containing the following keys:
        - namespace (str): The Kubernetes namespace for MetalLB.
        - repo (str): The URL of the MetalLB Helm chart repository.
        - version (str): The version of the MetalLB Helm chart.
        - values (str): The file path to the generated YAML values file.
        - release (str): The release name for the MetalLB deployment.
        - chart (str): The name of the MetalLB Helm chart.
    """

    metallb_values = {
        "namespace": "metallb-system",
        "chart_version": "0.14.9",
        "repo_url": "https://metallb.github.io/metallb",
        "chart_name": "metallb",
    }  # TODO: Change this to updated values

    Path(config_folder).joinpath(project_id, "metallb_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "metallb_values", "metallb_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(metallb_values))

    return {
        "namespace": metallb_values["namespace"],
        "repo": metallb_values["repo_url"],
        "version": metallb_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "metallb_values", "metallb_values.yaml")),
        "release": f"{project_id}-metallb",
        "chart": metallb_values["chart_name"],
    }


def create_cert_manager_values(config_folder, project_id):
    """
    Creates a dictionary of values for configuring cert-manager and writes it to a YAML file.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The project identifier.

    Returns
    -------
    dict
        A dictionary containing the namespace, repository URL, chart version, path to the values file, release name, and chart name.
    """

    cert_manager_chart_info = {
        "namespace": "cert-manager",
        "chart_version": "1.16.2",
        "repo_url": "https://charts.jetstack.io",
        "chart_name": "cert-manager",
    }

    cert_manager_values = {}
    cert_manager_values.update({"crds": {"enabled": True}})

    Path(config_folder).joinpath(project_id, "cert_manager_values").mkdir(parents=True, exist_ok=True)
    Path(config_folder).joinpath(project_id, "cert_manager_chart_info").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id, "cert_manager_values", "cert_manager_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(cert_manager_values))

    with open(Path(config_folder).joinpath(project_id, "cert_manager_chart_info", "cert_manager_chart_info.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(cert_manager_chart_info))

    return {
        "namespace": cert_manager_chart_info["namespace"],
        "repo": cert_manager_chart_info["repo_url"],
        "version": cert_manager_chart_info["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "cert_manager_values", "cert_manager_values.yaml")),
        "release": f"{project_id}-cert-manager",
        "chart": cert_manager_chart_info["chart_name"],
    }


def create_rancher_values(config_folder, project_id, cluster_config_dict):
    """
    Generates Rancher values configuration and writes it to a YAML file.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The project identifier.
    cluster_config_dict : dict
        A dictionary containing cluster configuration details.

    Returns
    -------
    dict
        A dictionary containing Rancher deployment details including namespace, repo URL,
        chart version, values file path, release name, and chart name.

    Example
    -------
    config_folder = "/path/to/config"
    project_id = "my_project"
    cluster_config_dict = {
        "traefik_resolver": "myresolver",
        "ingress_resolver_email": "email@example.com",
        "domain": "example.com",
        "rancher_password": "password"
    }
    create_rancher_values(config_folder, project_id, cluster_config_dict)
    """

    rancher_values = {
        "namespace": "cattle-system",
        "repo_url": "https://releases.rancher.com/server-charts/latest",
        "chart_name": "rancher",
        "chart_version": "2.10.1",
    }  # TODO: Change this to updated values

    rancher_values.update(
        {
            "hostname": "mgmt." + cluster_config_dict["domain"],
            "ingress": {"extraAnnotations": {}, "tls": {"source": "letsEncrypt"}},
            "letsEncrypt": {
                "email": cluster_config_dict["ingress_resolver_email"],
                "ingress": {"class": cluster_config_dict["ingress_class"]},
            },
            "bootstrapPassword": cluster_config_dict["rancher_password"],
        }
    )

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        rancher_values["ingress"]["extraAnnotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        rancher_values["ingress"]["extraAnnotations"]["traefik.ingress.kubernetes.io/router.tls"] = "true"
        rancher_values["ingress"]["extraAnnotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = (
            cluster_config_dict["traefik_resolver"]
        )
    elif cluster_config_dict["ingress_class"] == "nginx":
        rancher_values["ingress"]["extraAnnotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"

    Path(config_folder).joinpath(project_id, "rancher_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id, "rancher_values", "rancher_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(rancher_values))

    return {
        "namespace": rancher_values["namespace"],
        "repo": rancher_values["repo_url"],
        "version": rancher_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "rancher_values", "rancher_values.yaml")),
        "release": f"{project_id}-rancher",
        "chart": rancher_values["chart_name"],
    }


def create_gpu_operator_values(config_folder, project_id, cluster_config_dict):
    """
    Creates GPU operator values configuration for a Kubernetes cluster and writes it to a YAML file.

    Parameters
    ----------
    config_folder : str
        The folder path where the configuration will be saved.
    project_id : str
        The project identifier used to create a unique directory for the configuration.
    cluster_config_dict : dict
        A dictionary containing cluster configuration details, including the Kubernetes distribution.

    Returns
    -------
    dict
        A dictionary containing the namespace, repository URL, chart version, path to the values file, release name, and chart name.
    """

    gpu_operator_values = {
        "namespace": "gpu-operator",
        "chart_version": "25.3.1",
        "repo_url": "https://helm.ngc.nvidia.com/nvidia",
        "chart_name": "gpu-operator",
    }  # TODO: Change this to updated values

    if cluster_config_dict["k8s_distribution"] == "microk8s":
        gpu_operator_values["toolkit"] = {
            "env": [
                {"name": "CONTAINERD_CONFIG", "value": "/var/snap/microk8s/current/args/containerd-template.toml"},
                {"name": "CONTAINERD_SOCKET", "value": "/var/snap/microk8s/common/run/containerd.sock"},
                {"name": "CONTAINERD_RUNTIME_CLASS", "value": "nvidia"},
                {"name": "CONTAINERD_SET_AS_DEFAULT", "value": "true"},
            ]
        }

    elif cluster_config_dict["k8s_distribution"] == "rke2":
        gpu_operator_values["toolkit"] = {
            "driver": {"enabled": False},
            "env": [
                {"name": "CONTAINERD_SOCKET", "value": "/run/k3s/containerd/containerd.sock"},
                {"name": "CONTAINERD_CONFIG", "value": "/var/lib/rancher/rke2/agent/etc/containerd/config.toml.tmpl"},
                {"name": "CONTAINERD_RUNTIME_CLASS", "value": "nvidia"},
                {"name": "CONTAINERD_SET_AS_DEFAULT", "value": "true"},
            ],
        }

    Path(config_folder).joinpath(project_id, "gpu_operator_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id, "gpu_operator_values", "gpu_operator_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(gpu_operator_values))

    return {
        "namespace": gpu_operator_values["namespace"],
        "repo": gpu_operator_values["repo_url"],
        "version": gpu_operator_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "gpu_operator_values", "gpu_operator_values.yaml")),
        "release": f"{project_id}-gpu-operator",
        "chart": gpu_operator_values["chart_name"],
    }


def create_ingress_nginx_values(config_folder, project_id):
    """
    Creates and writes the ingress-nginx Helm chart values to a YAML file.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The unique identifier for the project.

    Returns
    -------
    dict
        A dictionary containing the following keys:
        - namespace (str): The namespace for the ingress-nginx.
        - repo (str): The repository URL for the ingress-nginx chart.
        - version (str): The version of the ingress-nginx chart.
        - values (str): The file path to the generated ingress-nginx values YAML file.
        - release (str): The release name for the ingress-nginx chart.
        - chart (str): The name of the ingress-nginx chart.
    """

    ingress_nginx_values = {
        "namespace": "ingress-nginx",
        "repo_url": "https://kubernetes.github.io/ingress-nginx",
        "chart_name": "ingress-nginx",
        "chart_version": "4.11.3",
    }  # TODO: Change this to updated values

    Path(config_folder).joinpath(project_id, "ingress_nginx_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id, "ingress_nginx_values", "ingress_nginx_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(ingress_nginx_values))

    return {
        "namespace": ingress_nginx_values["namespace"],
        "repo": ingress_nginx_values["repo_url"],
        "version": ingress_nginx_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "ingress_nginx_values", "ingress_nginx_values.yaml")),
        "release": f"{project_id}-ingress-nginx",
        "chart": ingress_nginx_values["chart_name"],
    }


def create_nfs_server_provisioner_values(config_folder, project_id, cluster_config_dict):
    """
    Creates and writes the NFS server provisioner Helm chart values to a YAML file.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The unique identifier for the project.
    cluster_config_dict : dict
        A dictionary containing cluster configuration details, including the NFS server and path.

    Returns
    -------
    dict
        A dictionary containing the following keys:
        - namespace (str): The namespace for the NFS server provisioner.
        - repo (str): The repository URL for the NFS server provisioner chart.
        - version (str): The version of the NFS server provisioner chart.
        - values (str): The file path to the generated NFS server provisioner values YAML file.
        - release (str): The release name for the NFS server provisioner chart.
        - chart (str): The name of the NFS server provisioner chart.
    """

    nfs_server_provisioner_values = {
        "namespace": "nfs-server-provisioner",
        "repo_url": "https://kubernetes-sigs.github.io/nfs-subdir-external-provisioner/",
        "chart_name": "nfs-subdir-external-provisioner",
        "chart_version": "4.0.18",
    }

    if "nfs_server" not in cluster_config_dict or "nfs_path" not in cluster_config_dict:
        nfs_server_provisioner_values.update({"nfs": {"server": "nfs-server.default.svc.cluster.local", "path": "/exports"}})
    else:
        nfs_server_provisioner_values.update(
            {"nfs": {"server": cluster_config_dict["nfs_server"], "path": cluster_config_dict["nfs_path"]}}
        )

    Path(config_folder).joinpath(project_id, "nfs_provisioner_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id, "nfs_provisioner_values", "nfs_provisioner_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(nfs_server_provisioner_values))

    return {
        "namespace": nfs_server_provisioner_values["namespace"],
        "repo": nfs_server_provisioner_values["repo_url"],
        "version": nfs_server_provisioner_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "nfs_provisioner_values", "nfs_provisioner_values.yaml")),
        "release": f"{project_id}-nfs-server-provisioner",
        "chart": nfs_server_provisioner_values["chart_name"],
    }


def create_gpu_booking_values(config_folder, project_id, cluster_config_dict, maia_config_dict):
    """
    Creates and writes GPU booking Helm chart values to a YAML file for a given project and cluster configuration.

    This function prepares a dictionary of values required for deploying the GPU booking Helm chart,
    including image repositories, API URLs, and authentication tokens. It writes these values to a YAML
    file in a structured directory under the specified configuration folder.

    Parameters
    ----------
    config_folder : str or Path
        The base directory where the configuration files should be stored.
    project_id : str
        The unique identifier for the project, used to create a subdirectory.
    cluster_config_dict : dict
        Dictionary containing cluster-specific configuration, must include the "domain" key.
    maia_config_dict : dict
        Dictionary containing MAIA-specific configuration, must include the "dashboard_api_secret" key.

    Returns
    -------
    dict
        A dictionary containing:
            - "namespace": The Kubernetes namespace for the deployment.
            - "repo": The Helm chart repository URL.
            - "version": The Helm chart version.
            - "values": The path to the generated YAML values file.
            - "release": The Helm release name.
            - "chart": The Helm chart name.

    Raises
    ------
    KeyError
        If required keys are missing from `cluster_config_dict` or `maia_config_dict`.
    OSError
        If there is an error creating directories or writing the YAML file.
    """
    gpu_booking_values = {
        "namespace": "maia-webhooks",
        "repo_url": "https://kthcloud.github.io/MAIA/",
        "chart_name": "gpu-booking",
        "chart_version": "1.0.0",
    }

    domain = cluster_config_dict["domain"]
    gpu_booking_values.update(
        {
            "image": {
                "pod_terminator": {
                    "repository": "kthcloud/gpu-booking-pod-terminator",
                    "pullPolicy": "IfNotPresent",
                    "tag": "1.4",
                },
                "repository": "kthcloud/gpu-booking-admission-controller",
                "pullPolicy": "IfNotPresent",
                "tag": "1.6",
            },
            "apiUrl": f"https://{domain}/maia-api/gpu-schedulability",
            "apiToken": maia_config_dict["dashboard_api_secret"],
        }
    )

    Path(config_folder).joinpath(project_id, "gpu_booking_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id, "gpu_booking_values", "gpu_booking_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(gpu_booking_values))

    return {
        "namespace": gpu_booking_values["namespace"],
        "repo": gpu_booking_values["repo_url"],
        "version": gpu_booking_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "gpu_booking_values", "gpu_booking_values.yaml")),
        "release": f"{project_id}-gpu-booking",
        "chart": gpu_booking_values["chart_name"],
    }



#!/usr/bin/env python

from __future__ import annotations

import asyncio
import datetime
import json
import os
import subprocess
from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from textwrap import dedent

import click
import hydra
import yaml
from hydra import compose as hydra_compose
from hydra import initialize_config_dir
from omegaconf import OmegaConf
from pyhelm3 import Client

import MAIA
from MAIA.maia_admin import (
    create_filebrowser_values,
    create_maia_namespace_values,
    generate_minio_configs,
    generate_mlflow_configs,
    generate_mysql_configs,
    get_maia_toolkit_apps,
    install_maia_project,
)
from MAIA.maia_fn import deploy_mlflow, deploy_mysql, deploy_oauth2_proxy, deploy_orthanc
from MAIA_scripts.MAIA_create_JupyterHub_config import create_jupyterhub_config_api

version = MAIA.__version__


TIMESTAMP = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.datetime.now())

DESC = dedent(
    """
    Script to deploy a MAIA Project Toolkit to a Kubernetes cluster. The target cluster is specified by setting
    the corresponding ``--cluster-config``, while the project-related configuration is specified with
    ``--project-config-file`` and ``--maia-config-file``. The necessary MAIA Configuration files
    should be found in ``--config-folder``.
    """  # noqa: E501
)
EPILOG = dedent(
    """
    Example call:
    ::
        {filename} --project-config-file /PATH/TO/form.yaml --cluster-config /PATH/TO/cluster.yaml
        --config-folder /PATH/TO/config_folder --maia-config-file /PATH/TO/maia_config.yaml
    """.format(  # noqa: E501
        filename=Path(__file__).stem
    )
)


def get_arg_parser():
    pars = ArgumentParser(description=DESC, epilog=EPILOG, formatter_class=RawTextHelpFormatter)

    pars.add_argument(
        "--project-config-file",
        type=str,
        required=True,
        help="YAML configuration file used to extract the project configuration.",
    )

    pars.add_argument(
        "--maia-config-file", type=str, required=True, help="YAML configuration file used to extract MAIA configuration."
    )

    pars.add_argument(
        "--cluster-config", type=str, required=True, help="YAML configuration file used to extract the cluster configuration."
    )

    pars.add_argument(
        "--config-folder",
        type=str,
        required=True,
        help="Configuration Folder where to locate (and temporarily store) the MAIA configuration files.",
    )

    pars.add_argument(
        "--minimal", type=str, required=False, help="Optional flag to only deploy JupyterHub in the MAIA namespace."
    )

    pars.add_argument("--no-argocd", action="store_true", help="Do not deploy with ArgoCD.")

    pars.add_argument("-v", "--version", action="version", version="%(prog)s " + version)

    return pars


async def verify_installed_maia_toolkit(project_id, namespace, get_chart_metadata=True):

    client = Client(kubeconfig=os.environ["KUBECONFIG"])

    try:
        revision = await client.get_current_revision(project_id, namespace=namespace)
    except Exception as e:
        print(f"An error occurred: {e}")
        print("Project not found")
        return -1
    if get_chart_metadata:
        chart_metadata = await revision.chart_metadata()
        print(
            revision.release.name,
            revision.release.namespace,
            revision.revision,
            str(revision.status),
            chart_metadata.name,
            chart_metadata.version,
        )
        return revision.revision
    return {
        "revision": revision.revision,
        "release_name": revision.release.name,
        "release_namespace": revision.release.namespace,
        "status": str(revision.status),
    }


@click.command()
@click.option("--project-config-file", type=str)
@click.option("--maia-config-file", type=str)
@click.option("--cluster-config", type=str)
@click.option("--no-minimal", is_flag=True)
@click.option("--config-folder", type=str)
@click.option("--no-argocd", is_flag=True)
def main(project_config_file, maia_config_file, cluster_config, config_folder, no_minimal=False, no_argocd=False):
    deploy_maia_toolkit(project_config_file, maia_config_file, cluster_config, config_folder, no_minimal, no_argocd)


def deploy_maia_toolkit(project_config_file, maia_config_file, cluster_config, config_folder, no_minimal=False, no_argocd=False):
    project_form_dict = yaml.safe_load(Path(project_config_file).read_text())

    cluster_config_dict = yaml.safe_load(Path(cluster_config).read_text())
    maia_config_dict = yaml.safe_load(Path(maia_config_file).read_text())

    deploy_maia_toolkit_api(project_form_dict, maia_config_dict, cluster_config_dict, config_folder, not no_minimal, no_argocd)


def deploy_maia_toolkit_api(
    project_form_dict,
    maia_config_dict,
    cluster_config_dict,
    config_folder,
    minimal=True,
    no_argocd=False,
    redeploy_enabled=True,
    return_values_only=False,
):
    project_form_dict["extra_configs"] = {"enable_cifs": True}  # Default value, can be overridden by the form
    private_maia_registry = os.environ.get("MAIA_PRIVATE_REGISTRY", None)
    group_id = project_form_dict["group_ID"]
    Path(config_folder).joinpath(project_form_dict["group_ID"]).mkdir(parents=True, exist_ok=True)

    namespace = project_form_dict["group_ID"].lower().replace("_", "-")

    helm_commands = []

    mlflow_configs = generate_mlflow_configs(namespace=group_id.lower().replace("_", "-"))

    if not minimal:
        minio_configs = generate_minio_configs(namespace=group_id.lower().replace("_", "-"))

        mysql_configs = generate_mysql_configs(namespace=group_id.lower().replace("_", "-"))

        project_form_dict["minio_access_key"] = minio_configs["console_access_key"]
        project_form_dict["minio_secret_key"] = minio_configs["console_secret_key"]
    else:
        minio_configs = None
        mysql_configs = None
        project_form_dict["minio_access_key"] = "N/A"
        project_form_dict["minio_secret_key"] = "N/A"

    print("Creating MAIA Project Namespace")
    if no_argocd:
        print("No ArgoCD deployment")
    else:
        print("ArgoCD deployment")

    helm_commands.append(
        create_maia_namespace_values(
            project_form_dict, cluster_config_dict, config_folder, minio_configs=minio_configs, mlflow_configs=mlflow_configs
        )
    )

    with open(Path(config_folder).joinpath(group_id, "maia_namespace_values", "namespace_values.yaml")) as f:
        maia_namespace_values = yaml.safe_load(f)
        project_form_dict["ssh_users"] = []
        for user in maia_namespace_values["users"]:
            project_form_dict["ssh_users"].append(
                {
                    "username": user["jupyterhub_username"].replace("-2d", "-").replace("-40", "@").replace("-2e", "."),
                    "ssh_port": user["sshPort"],
                }
            )

    helm_commands.append(
        create_jupyterhub_config_api(project_form_dict, maia_config_dict, cluster_config_dict, config_folder, minimal=minimal)
    )

    helm_commands.append(
        create_filebrowser_values(project_form_dict, cluster_config_dict, config_folder, mlflow_configs=mlflow_configs)
    )
    if not minimal:
        helm_commands.append(deploy_oauth2_proxy(cluster_config_dict, project_form_dict, config_folder))

        helm_commands.append(deploy_mysql(cluster_config_dict, project_form_dict, config_folder, mysql_configs=mysql_configs))
        helm_commands.append(
            deploy_mlflow(
                cluster_config_dict,
                project_form_dict,
                config_folder,
                maia_config_dict,
                mysql_config=mysql_configs,
                minio_config=minio_configs,
            )
        )

        helm_commands.append(deploy_orthanc(cluster_config_dict, project_form_dict, maia_config_dict, config_folder))

    json_key_path = os.environ.get("JSON_KEY_PATH", None)
    for helm_command in helm_commands:
        if not helm_command["repo"].startswith("http") and not return_values_only:
            original_repo = helm_command["repo"]
            helm_command["repo"] = f"oci://{helm_command['repo']}"
            try:
                with open(json_key_path, "r") as f:
                    docker_credentials = json.load(f)
                    username = docker_credentials.get("harbor_username")
                    password = docker_credentials.get("harbor_password")
            except Exception:
                with open(json_key_path, "r") as f:
                    docker_credentials = f.read()
                    username = "_json_key"
                    password = docker_credentials

            subprocess.run(
                ["helm", "registry", "login", original_repo, "--username", username, "--password-stdin"], stdin=password.encode()
            )

            print(" ".join(["helm", "registry", "login", original_repo, "--username", username, "--password-stdin"]))
            subprocess.run(
                [
                    "helm",
                    "pull",
                    helm_command["repo"] + "/" + helm_command["chart"],
                    "--version",
                    helm_command["version"],
                    "--destination",
                    "/tmp",
                ]
            )
            print(
                " ".join(
                    [
                        "helm",
                        "pull",
                        helm_command["repo"] + "/" + helm_command["chart"],
                        "--version",
                        helm_command["version"],
                        "--destination",
                        "/tmp",
                    ]
                )
            )

            cmd = [
                "helm",
                "upgrade",
                "--install",
                # "--wait",
                "-n",
                helm_command["namespace"],
                helm_command["release"],
                "/tmp/" + helm_command["chart"] + "-" + helm_command["version"] + ".tgz",
                "--values",
                helm_command["values"],
            ]
            print(" ".join(cmd))
        else:
            cmd = [
                "helm",
                "upgrade",
                "--install",
                # "--wait",
                "-n",
                helm_command["namespace"],
                helm_command["release"],
                helm_command["chart"],
                "--repo",
                helm_command["repo"],
                "--version",
                helm_command["version"],
                "--values",
                helm_command["values"],
            ]
            print(" ".join(cmd))
        if no_argocd and not return_values_only:

            subprocess.Popen(cmd)

    destination_cluster_address = cluster_config_dict["argocd_destination_cluster_address"]

    values = {
        "defaults": [
            "_self_",
            {"maia_namespace_values": "namespace_values"},
            {"jupyterhub_values": "jupyterhub_values"},
            {"jupyterhub_chart_info": "jupyterhub_chart_info"},
            {"maia_filebrowser_values": "maia_filebrowser_values"},
        ],
        "argo_namespace": maia_config_dict["argocd_namespace"],
        "group_ID": f"MAIA:{group_id}",
        "destination_server": f"{destination_cluster_address}",
        "sourceRepos": [
            "https://kthcloud.github.io/MAIA/",
            "https://hub.jupyter.org/helm-chart/",
            "https://oauth2-proxy.github.io/manifests",
        ],
    }
    if private_maia_registry:
        values["sourceRepos"].append(private_maia_registry)
    if not minimal:
        values["defaults"].append({"mlflow_values": "mlflow_values"})
        values["defaults"].append({"mysql_values": "mysql_values"})
        values["defaults"].append({"oauth2_proxy_values": "oauth2_proxy_values"})
        values["defaults"].append({"orthanc_values": "orthanc_values"})

    with open(Path(config_folder).joinpath(group_id, "values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(values))

    try:
        initialize_config_dir(config_dir=str(Path(config_folder).joinpath(group_id)), job_name=group_id)
    except Exception as e:
        print(f"An error occurred: {e}")
        hydra.core.global_hydra.GlobalHydra.instance().clear()
        initialize_config_dir(config_dir=str(Path(config_folder).joinpath(group_id)), job_name=group_id)
    cfg = hydra_compose("values.yaml")
    OmegaConf.save(cfg, str(Path(config_folder).joinpath(group_id, f"{group_id}_values.yaml")), resolve=True)

    project_id = namespace

    if no_argocd:
        if return_values_only:
            with open(Path(config_folder).joinpath(group_id, f"{group_id}_values.yaml")) as f:
                return yaml.safe_load(f)
        else:
            return
    revision = asyncio.run(verify_installed_maia_toolkit(project_id, maia_config_dict["argocd_namespace"]))

    if revision == -1 or redeploy_enabled:
        print("Installing MAIA Workspace")

        project_chart = maia_config_dict["maia_project_chart"]
        project_repo = maia_config_dict["maia_project_repo"]
        project_version = maia_config_dict["maia_project_version"]
        json_key_path = None
        if not minimal:
            project_chart = maia_config_dict["maia_pro_project_chart"]
            project_repo = maia_config_dict["maia_pro_project_repo"]
            project_version = maia_config_dict["maia_pro_project_version"]
            json_key_path = os.environ.get("JSON_KEY_PATH", None)
        print(f"Installing MAIA Project Toolkit {project_chart} from {project_repo} version {project_version}")
        print(f"Using JSON key path: {json_key_path}")
        msg = asyncio.run(
            install_maia_project(
                group_id,
                Path(config_folder).joinpath(group_id, f"{group_id}_values.yaml"),
                maia_config_dict["argocd_namespace"],
                project_chart,
                project_repo=project_repo,
                project_version=project_version,
                json_key_path=json_key_path,
            )
        )
        return msg
    else:
        argocd_host = maia_config_dict["argocd_host"]
        token = maia_config_dict["argocd_token"]
        print("MAIA Workspace already installed")
        asyncio.run(get_maia_toolkit_apps(group_id, token, argocd_host))


# The `if __name__ == "__main__":` block in Python is used to check whether the current script is
# being run directly by the Python interpreter or if it is being imported as a module into another
# script.
if __name__ == "__main__":
    main()
